<'
# Copyright (c) 2011-2015,2017,2024 LAAS/CNRS
# All rights reserved.
#
# Redistribution and use  in source  and binary  forms,  with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   1. Redistributions of  source  code must retain the  above copyright
#      notice and this list of conditions.
#   2. Redistributions in binary form must reproduce the above copyright
#      notice and  this list of  conditions in the  documentation and/or
#      other materials provided with the distribution.
#
#                                      Anthony Mallet on Wed Feb 29 2012
#

if {[llength $argv] != 1} { error "expected arguments: component" }
lassign $argv component

# compute handy shortcuts
set comp [$component name]
set COMP [string toupper [$component name]]

lang c
'>
#ifndef H_<"$COMP">_CONTROL_TASK
#define H_<"$COMP">_CONTROL_TASK

#include <assert.h>
#include <pthread.h>
#include <stdarg.h>

#include "csLib.h"
#include "semLib.h"

#include "<"$comp">_spinlock.h"
#include "<"$comp">_c_types.h"
#include "<"$comp">_parameters.h"
#include "<"$comp">_msglib.h"
#include "<"$comp">_portlib.h"
#include "<"$comp">_remotelib.h"
#include "<"$comp">_activity.h"


/* --- internal state ------------------------------------------------------ */

extern const char *genom_instance;

struct genom_component_data;

struct genom_context_data {
  struct genom_component_data *self;

  genom_event ex;
  void *exdetail;
  size_t exsize;
};

struct genom_component_data {
<'if {![catch {$component ids} ids]} {'>
  <"[$ids declarator ids]">;
<'}'>

  struct {
<'foreach p [$component ports] {'>
    struct genom_<"$comp">_<"[$p name]">_port <"[$p name]">;
<'}'>
  } ports;

  struct {
<'foreach r [$component remotes] {'>
    struct genom_<"$comp">_<"[$r name]">_remote <"[$r name]">;
<'}'>
  } remotes;

  struct {
<'foreach t [$component tasks] {'>
    struct {
      struct genom_context_iface context;
      struct genom_context_data context_data;
      long taskid;

      genom_state_rusage rusage;
      pthread_spinlock_t rlock;

      int runnable, shutdown;
      struct genom_activity *permanent;
      struct genom_activities activities;
      pthread_mutex_t lock;
      pthread_cond_t sync;
    } <"[$t name]">;
<'}'>
  } tasks;

  /* control task */
  struct {
    struct genom_context_iface context;
    struct genom_context_data context_data;
    long taskid;

    char mbox_name[64];
    SERV_ID csserv;
    int shutdown;

    struct timeval rusage_update;
    pthread_spinlock_t rlock;

    union genom_any_activity activity;
    int run_map[<"$COMP">_NRQSTID]; /* before/after constraints */
    pthread_mutex_t lock;
    pthread_cond_t sync;
  } control;

  /* mutual exclusion for resource access */
  struct {
    int all;				/* all resources locked */
    void *control;			/* control task codel/service */
<'foreach t [$component tasks] {'>
    void *task_<"[$t name]">;		/* <"[$t name]"> codel */
<'}'>

    unsigned int q, qnext;		/* fifo (ticket lock) */
    pthread_mutex_t lock;
    pthread_cond_t sync;
  } resources;
};

/* resource access */
#define genom_take_resource(self, wait, set) do {                       \
    pthread_mutex_lock(&(self)->resources.lock);                        \
    unsigned int q = (self)->resources.qnext++;                         \
    while(q != self->resources.q || (wait))                             \
      pthread_cond_wait(&(self)->resources.sync,                        \
                        &(self)->resources.lock);                       \
    set;                                                                \
    (self)->resources.q++;                                              \
    pthread_cond_broadcast(&self->resources.sync);                      \
    pthread_mutex_unlock(&self->resources.lock);                        \
  } while(0)

#define genom_give_resource(self, set) do {                             \
    pthread_mutex_lock(&(self)->resources.lock);                        \
    set;                                                                \
    pthread_cond_broadcast(&self->resources.sync);                      \
    pthread_mutex_unlock(&self->resources.lock);                        \
  } while(0)


/* control task */
void *	genom_<"$comp">_init(void);
void	genom_<"$comp">_fini(void *data);
void	genom_<"$comp">_schedule_cntrl(struct genom_component_data *self);

<'foreach s [$component services] {'>
genom_event genom_<"$comp">_<"[$s name]">_controlcb(
   struct genom_component_data *,
   struct genom_<"$comp">_<"[$s name]">_activity *);
<'}'>


/* exec tasks */
<'foreach t [$component tasks] {'>
void *		genom_<"$comp">_<"[$t name]">_exec_task(void *data);
void		genom_<"$comp">_<"[$t name]">_invoke_perm(
			struct genom_component_data *);
enum genom_activity_status
		genom_<"$comp">_<"[$t name]">_invoke(
			struct genom_component_data *,
			struct genom_activity *);
<'}'>


/* exceptions */
genom_event	genom_pocolibs_raise(genom_event ex, const void *detail,
                        size_t size, genom_context self);
const void *	genom_pocolibs_raised(genom_event *ex, genom_context self);


/* log functions */
void	genom_log_info(const char *format, ...)
  __attribute__ ((format (printf, 1, 2)));
void	genom_log_warn(int h2error, const char *format, ...)
  __attribute__ ((format (printf, 2, 3)));
void	genom_log_debug(const char *format, ...)
  __attribute__ ((format (printf, 1, 2)));


#endif /* H_<"$COMP">_CONTROL_TASK */
